#include <uart/dma-ring-buffer.h>
#include <clock-enable-manager.h>
#include <stdbool.h>

static int dma_ring_buffer_switch_clock_enable(uint8_t base_dma, bool clk_en)
{
	int ret_val;
	int (*clk_func)(volatile uint32_t *, uint8_t);

	if (clk_en)
		clk_func = rcc_manager_enable_clock;
	else
		clk_func = rcc_manager_disable_clock;

	switch (base_dma) {
	case 1:
		ret_val = clk_func(&RCC->AHB1ENR, BITMASK_TO_BITNO(RCC_AHB1ENR_DMA1EN));
		break;
	case 2:
		ret_val = clk_func(&RCC->AHB1ENR, BITMASK_TO_BITNO(RCC_AHB1ENR_DMA2EN));
		break;
	default:
		ret_val = -1000;
		break;
	}

	return ret_val;
}

int dma_ring_buffer_periph_to_mem_initialize(struct dma_ring_buffer_to_mem *dma_buffer, uint8_t base_dma_id,
					     DMA_Stream_TypeDef *dma_stream, size_t buffer_element_count, size_t element_size,
					     void *data_buffer, void* src_reg, uint8_t dma_trigger_channel)
{
	int ret_val = 0;

	if (!dma_buffer || !dma_stream || !data_buffer || !src_reg)
		return -1000;

	if (dma_trigger_channel > 7)
		return -1007;

	dma_buffer->base_dma_id = base_dma_id;

	ret_val = dma_ring_buffer_switch_clock_enable(base_dma_id, true);
	if (ret_val)
		return ret_val;

	dma_buffer->dma = dma_stream;
	dma_buffer->get_idx = 0;
	dma_buffer->buffer_count = buffer_element_count;

	dma_buffer->data_ptr = data_buffer;
	dma_buffer->element_size = element_size;

	dma_stream->PAR = (uint32_t)src_reg;
	dma_stream->M0AR = (uint32_t)data_buffer;
	dma_stream->NDTR = buffer_element_count;
	dma_stream->NDTR = buffer_element_count;

	dma_stream->CR |= (dma_trigger_channel<<25) | DMA_SxCR_MINC | DMA_SxCR_CIRC | DMA_SxCR_EN;

	return 0;
}

int dma_ring_buffer_periph_to_mem_get_data(struct dma_ring_buffer_to_mem *buff, const void **data_buff, size_t *len)
{
	int ret_code = 0;
	uint32_t ndtr;
	size_t put_idx;

	if (!buff || !data_buff || !len)
		return -1;


	ndtr = buff->dma->NDTR;
	put_idx = buff->buffer_count - ndtr;

	/* Check if wrap around */
	if (put_idx < buff->get_idx) {
		*data_buff = &(((char *)buff->data_ptr)[buff->get_idx * buff->element_size]);
		*len = buff->buffer_count - buff->get_idx;
		buff->get_idx = 0;
		ret_code = 1;
	} else if (put_idx > buff->get_idx) {
		*data_buff = &(((char *)buff->data_ptr)[buff->get_idx * buff->element_size]);
		*len = put_idx - buff->get_idx;
		buff->get_idx += *len;
	} else {
		/* No new data */
		*len = 0;
	}

	return ret_code;
}

void dma_ring_buffer_periph_to_mem_stop(struct dma_ring_buffer_to_mem *buff)
{
	if (!buff || !buff->dma)
		return;

	buff->dma->CR = 0;
	buff->dma->NDTR = 0;
	buff->dma->M1AR = 0;
	buff->dma->FCR = 0;

	dma_ring_buffer_switch_clock_enable(buff->base_dma_id, false);
}
